@@ -1,10 +1,22 @@
 import time
 
-from .compat import jwt_encode
+# from typing import Protocol
+from typing import Dict
+
+import jwt
 
 JWT_ALG = "HS256"
 
 
+HeaderDict = Dict[str, str]
+
+
+# class ClientAuthProtocol(Protocol):
+
+#     def credentials(self) -> HeaderDict:
+#         ...
+
+
 class ClientAuth:
     """
     Auth for the ZDS client, using JWT.
@@ -22,6 +34,8 @@ class ClientAuth:
         'Authorization': '<base64>.<base64>.<base64>'
     }
     >>> requests.get(url, **auth.credentials())
+
+    TODO: make pluggable and implement some common auth schemes.
     """
 
     def __init__(
@@ -60,7 +74,7 @@ class ClientAuth:
                 **self.claims,
             }
 
-            encoded = jwt_encode(payload, self.secret, algorithm=JWT_ALG)
+            encoded = jwt.encode(payload, self.secret, algorithm=JWT_ALG)
 
             self._credentials = {
                 "Authorization": "Bearer {encoded}".format(encoded=encoded)
