@@ -1,18 +1,14 @@
-import copy
 import logging
 import re
-import warnings
-from typing import Any, Dict, List, Optional, Union
+from dataclasses import dataclass, field
+from typing import Any, Callable, Dict, List, Optional, Union
 from urllib.parse import urljoin, urlparse
 
 import requests
-import yaml
 from requests.structures import CaseInsensitiveDict
 
-from .config import ClientConfig
-from .log import Log
+from .auth import ClientAuth
 from .oas import schema_fetcher
-from .registry import registry
 from .schema import get_headers, get_operation_url
 
 logger = logging.getLogger(__name__)
@@ -29,14 +25,8 @@ class ClientError(Exception):
     pass
 
 
-class Client:
-
-    _schema = None
-    _log = Log()
-
-    auth = None
-
-    operation_suffix_mapping = {
+def _get_default_op_suffix_mapping() -> Dict[str, str]:
+    return {
         "list": "_list",
         "retrieve": "_read",
         "create": "_create",
@@ -45,126 +35,108 @@ class Client:
         "delete": "_delete",
     }
 
-    def __init__(self, service: str, base_path: str = "/api/v1/"):
-        """
-        Obtain a client instance from the configuration registry.
 
-        :param service: Alias of the service/API
+def _get_session(client: Optional["Client"] = None) -> requests.Session:
+    """
+    Get a session to leverage connection pooling.
+
+    TODO: this is work in progress
+    """
+    # TODO: cache session on the client instance if provided
+    session = requests.Session()
+    if client and client.request_hooks:
+        session.hooks.update(client.request_hooks)
+    return session
+
+
+@dataclass
+class Client:
+    api_root: str
+    """
+    Fully qualified base URL of the API/service, e.g. ``"https://example.com/api/v1/"``.
+    """
+    oas_location: str = ""
+    """
+    Relative location of the OpenAPI spec.
+
+    By default, it is assumed the API spec is hosted at the API root.
+    """
+    auth: Optional[ClientAuth] = None
+    """
+    Optional :class:`ClientAuth`-like interface specifying the authentication parameters.
+    """
+    operation_suffix_mapping: Dict[str, str] = field(
+        default_factory=_get_default_op_suffix_mapping
+    )
+    """
+    Mapping of semantic RESTful operations to the operation-IDs used in the API spec.
+
+    TODO: make this more robust - operationId is an _optional_ key and may be empty.
+    """
+    request_hooks: Optional[Dict[str, Callable[..., Any]]] = None
+    """
+    requests library hooks to apply for every request.
+
+    See https://requests.readthedocs.io/en/latest/user/advanced/#event-hooks for
+    details. You can use this to set up request logging or middleware-like processing
+    to the responses.
+    """
+
+    _schema: Optional[dict] = field(init=False, default=None)
+
+    def __post_init__(self):
+        # normalize the API root - this should always be used as base URL
+        # for relative URLs
+        if not self.api_root.endswith("/"):
+            self.api_root = f"{self.api_root}/"
+
+    def _build_url(self, url_like: str):
         """
-        try:
-            self._config = registry[service]
-        except KeyError:
-            raise RuntimeError(
-                "Service '{service}' is not known in the client registry. "
-                "Did you load the config first through `Client.load_config(path, **manual)`?".format(
-                    service=service
-                )
+        Build a fully qualified URL based on the API root.
+
+        If the URL is fully qualified, return it unmodified. Otherwise, assert that it's
+        a relative path and join it with the API root. If it's not relative but absolute
+        (starting with a slash), emit a debug message.
+        """
+        if url_like.startswith("/"):
+            logger.debug(
+                "Received an absolute path '%s', this may not be what you intended. "
+                "Absolute paths discard the path portion of the API configured root.",
+                url_like,
             )
+            # do not silently correct or strip this, it *may* be intentional
+            return urljoin(self.api_root, url_like)
 
-        self.service = service
-        self.base_path = base_path
+        # check if it's a fully qualified URL
+        parsed_url = urlparse(url_like)
+        if parsed_url.scheme and parsed_url.netloc and parsed_url.path:
+            return url_like
+        return urljoin(self.api_root, url_like)
 
-        self._base_url = None
-
-        self.auth = self._config.auth
-
-    def __repr__(self):
-        return "<%s: service=%r base_url=%r>" % (
-            self.__class__.__name__,
-            self.service,
-            self.base_url,
-        )
-
-    @classmethod
-    def load_config(cls, path: str = None, **manual):
-        """
-        Initialize the client configuration.
-
-        The configuration is delegated to the registry, so multiple instances
-        of the client share the same configuration.
-
-        The config file should have the following format:
-
-        .. code-block:: yaml
-
-            alias1:
-              scheme: http(s)
-              host: localhost
-              port: 8000
-              auth:
-                client_id: some-client-id
-                secret: very-secret
-
-        Multiple service configs are supported, each with their own alias.
-        The `port` and `auth` keys are optional. Port will default to 80 or
-        443 depending on the scheme.
-
-        :param path: path to the yaml file holding the config
-        :param manual: any manual overrides, as kwargs. Note this completely
-          overwrites any existing config in the YAML file if specified.
-        """
-        if path is not None:
-            logger.info("Loading config from %s", path)
-            with open(path, "r") as config_file:
-                client_configs = yaml.safe_load(config_file)
-
-            for alias, _config in client_configs.items():
-                config = ClientConfig.from_dict(_config)
-                registry.register(alias, config)
-
-        if manual:
-            logger.info("Applying manual config: %r", manual)
-            for alias, _config in manual.items():
-                config = ClientConfig.from_dict(_config)
-                registry.register(alias, config)
-
-    @classmethod
-    def from_url(cls, detail_url: str) -> "Client":
-        parsed_url = urlparse(detail_url)
-
-        # we know that API endpoints look like:
-        # - /base_path/collection/<uuid> or
-        # - /base_path/collection/<uuid>/subcollection or
-        # - /base_path/collection/<uuid>/subcollection/<uuid>
-        # So, splitting on UUIDs gives us the base_path + collection
-        bits = re.split(UUID_PATTERN, parsed_url.path)
-        base_path = (bits[0].rstrip("/").rsplit("/", 1))[0] + "/"
-
-        # register the config
-        config = ClientConfig.from_url(detail_url)
-        alias = config.base_url
-        registry.register(alias, config)
-        return cls(alias, base_path)
+    # OpenAPI schema fetching/loading
 
     @property
-    def log(self):
-        """
-        Local log entries.
-        """
-        return (
-            entry for entry in self._log.entries() if entry["service"] == self.service
-        )
-
-    @property
-    def base_url(self) -> str:
-        default = f"{self._config.base_url}{self.base_path}"
-        return self._base_url or default
-
-    @base_url.setter
-    def base_url(self, base_url: str) -> None:
-        self._base_url = base_url
-
-    @property
-    def schema(self):
+    def schema(self) -> dict:
         if self._schema is None:
             self.fetch_schema()
-        return self._schema
+        return self._schema or {}
 
-    def pre_request(self, method: str, url: str, **kwargs) -> Any:
+    @schema.setter
+    def schema(self, value: dict):
+        self._schema = value
+
+    def fetch_schema(self) -> None:
+        url = self._build_url(self.oas_location)
+        logger.info("Fetching schema at '%s'", url)
+        self._schema = schema_fetcher.fetch(url)
+
+    # Making requests for resources
+
+    def pre_request(self, method: str, url: str, kwargs: dict = None) -> Any:
         """
         Perform any pre-request processing required.
 
-        The kwargs are literally passed to :meth:`requests.request` and may
+        The kwargs are literally passed to the underlying requests call and may
         be mutated in place.
 
         The return value is passed as first argument to :meth:`post_response`.
@@ -179,7 +151,7 @@ class Client:
         expected_status=200,
         request_kwargs: Optional[dict] = None,
         **kwargs,
-    ) -> Union[List[Object], Object]:
+    ) -> Union[List[Object], Object, None]:
         """
         Make the HTTP request using requests.
 
@@ -190,7 +162,7 @@ class Client:
         :raises: :class:`requests.HTTPException` for internal server errors
         :raises: :class:`ClientError` for HTTP 4xx status codes
         """
-        url = urljoin(self.base_url, path)
+        url = urljoin(self.api_root, path)
 
         if request_kwargs:
             kwargs.update(request_kwargs)
@@ -206,9 +178,10 @@ class Client:
 
         kwargs["headers"] = headers
 
-        pre_id = self.pre_request(method, url, **kwargs)
+        pre_id = self.pre_request(method, url, kwargs)
 
-        response = requests.request(method, url, **kwargs)
+        with _get_session(self) as session:
+            response = session.request(method, url, **kwargs)
 
         try:
             response_json = response.json()
@@ -217,18 +190,6 @@ class Client:
 
         self.post_response(pre_id, response_json)
 
-        self._log.add(
-            self.service,
-            url,
-            method,
-            dict(headers),
-            copy.deepcopy(kwargs.get("data", kwargs.get("json", None))),
-            response.status_code,
-            dict(response.headers),
-            response_json,
-            params=kwargs.get("params"),
-        )
-
         try:
             response.raise_for_status()
         except requests.HTTPError as exc:
@@ -244,31 +205,20 @@ class Client:
     ) -> None:
         pass
 
-    def fetch_schema(self) -> None:
-        url = urljoin(self.base_url, "schema/openapi.yaml")
-        logger.info("Fetching schema at '%s'", url)
-        self._schema = schema_fetcher.fetch(url, {"v": "3"})
+    # Convenience/semantic wrappers
 
     def list(
         self,
         resource: str,
         params=None,
-        query_params=None,
         request_kwargs: Optional[dict] = None,
         **path_kwargs,
     ) -> List[Object]:
         op_suffix = self.operation_suffix_mapping["list"]
         operation_id = f"{resource}{op_suffix}"
         url = get_operation_url(
-            self.schema, operation_id, base_url=self.base_url, **path_kwargs
+            self.schema, operation_id, base_url=self.api_root, **path_kwargs
         )
-        if query_params and not params:
-            warnings.warn(
-                "Client.list 'query_params' kwarg is deprecated, use 'params' instead.",
-                DeprecationWarning,
-            )
-            params = query_params
-
         return self.request(
             url, operation_id, params=params, request_kwargs=request_kwargs
         )
@@ -284,7 +234,7 @@ class Client:
         operation_id = f"{resource}{op_suffix}"
         if url is None:
             url = get_operation_url(
-                self.schema, operation_id, base_url=self.base_url, **path_kwargs
+                self.schema, operation_id, base_url=self.api_root, **path_kwargs
             )
         return self.request(url, operation_id, request_kwargs=request_kwargs)
 
@@ -298,7 +248,7 @@ class Client:
         op_suffix = self.operation_suffix_mapping["create"]
         operation_id = f"{resource}{op_suffix}"
         url = get_operation_url(
-            self.schema, operation_id, base_url=self.base_url, **path_kwargs
+            self.schema, operation_id, base_url=self.api_root, **path_kwargs
         )
         return self.request(
             url,
@@ -321,7 +271,7 @@ class Client:
         operation_id = f"{resource}{op_suffix}"
         if url is None:
             url = get_operation_url(
-                self.schema, operation_id, base_url=self.base_url, **path_kwargs
+                self.schema, operation_id, base_url=self.api_root, **path_kwargs
             )
         return self.request(
             url,
@@ -344,7 +294,7 @@ class Client:
         operation_id = f"{resource}{op_suffix}"
         if url is None:
             url = get_operation_url(
-                self.schema, operation_id, base_url=self.base_url, **path_kwargs
+                self.schema, operation_id, base_url=self.api_root, **path_kwargs
             )
         return self.request(
             url,
@@ -366,7 +316,7 @@ class Client:
         operation_id = f"{resource}{op_suffix}"
         if url is None:
             url = get_operation_url(
-                self.schema, operation_id, base_url=self.base_url, **path_kwargs
+                self.schema, operation_id, base_url=self.api_root, **path_kwargs
             )
         return self.request(
             url,
@@ -385,9 +335,10 @@ class Client:
         request_kwargs: Optional[dict] = None,
         **path_kwargs,
     ) -> Union[List[Object], Object]:
+        # TODO: support expected_status kwarg
         if url is None:
             url = get_operation_url(
-                self.schema, operation_id, base_url=self.base_url, **path_kwargs
+                self.schema, operation_id, base_url=self.api_root, **path_kwargs
             )
         return self.request(
             url, operation_id, method=method, json=data, request_kwargs=request_kwargs
